import tempfile


def _load_xml(ps):
    with open(str(ps.raw_path), 'w'):
        pass


def _parse_xml(ps):
    with open(str(ps.parsed_path), 'w'):
        pass


def test_manager_workspace_update():
    from woger import WorkspaceManager, BaseAction, BasePathStructure

    class PathStructure(BasePathStructure):
        raw_path = 'data.xml'
        parsed_path = 'data.json'

    class Action(BaseAction):
        load_xml = 'load_data'
        parse_xml = 'parse_xml'

    with tempfile.TemporaryDirectory() as root:

        manager = WorkspaceManager(root, PathStructure)
        assert manager.current() is None

        latest_finished_ws = manager.find_latest_finished(Action.parse_xml)
        assert latest_finished_ws is None

        timestamp = 1346134
        manager.create(timestamp)
        manager.target_latest()

        ws = manager.target()
        tracker = ws.track(Action.load_xml)

        if not tracker.finished():
            assert tracker.started() or tracker.failed() or tracker.undefined()
            with tracker:
                assert tracker.started()
                _load_xml(ws.path)
            assert tracker.finished()

        tracker = ws.track(Action.parse_xml)

        if not tracker.finished():
            with tracker:
                assert tracker.started()
                _parse_xml(ws.path)
            assert tracker.finished()

        assert manager.current() is None
        manager.update()
        assert manager.current() == manager.storage[timestamp] == manager.target()



import enum
import tempfile

import pytest


class _ParseException(Exception):
    pass


def _body(ws, Action):

    load_tracker = ws.track(Action.load_raw_xml)
    with load_tracker:
        assert load_tracker.started()
        assert not load_tracker.finished()
        assert not load_tracker.failed()

    assert load_tracker.finished()
    assert not load_tracker.started()
    assert not load_tracker.failed()

    parse_tracker = ws.track(Action.parse_xml)
    with pytest.raises(_ParseException):
        with parse_tracker:
            assert parse_tracker.started()
            assert not parse_tracker.failed()
            assert not parse_tracker.finished()
            assert load_tracker.finished()
            raise _ParseException()

    assert parse_tracker.failed()
    assert not parse_tracker.finished()
    assert not parse_tracker.started()
    assert load_tracker.finished()
    assert not load_tracker.started()
    assert not load_tracker.failed()


def test_manager_state_updates():
    from woger import Workspace, BaseAction, BasePathStructure

    @enum.unique
    class Action(BaseAction):
        load_raw_xml = 'load_raw_xml'
        parse_xml = 'parse_xml'

    class PathStructure(BasePathStructure):
        pass

    with tempfile.TemporaryDirectory() as root:
        ws = Workspace(PathStructure(root))
        _body(ws, Action)



import warnings
from operator import attrgetter, itemgetter
from typing import List, Optional, Iterable

import os

from .workspace import Workspace


def _load_workspaces(root_path,
                     path_structure_cls,
                     data_cls,
                     workspace_cls):
    path_list = [
        root_path / name
        for name
        in os.listdir(str(root_path))
    ]
    filtered = [
        path for path in path_list
        if os.path.isdir(str(path))
    ]
    return [
        workspace_cls(path_structure_cls(path), data_cls)
        for path
        in filtered
    ]


class WorkspaceStorage:
    """Stores workspaces"""

    def __init__(self, workspaces: Optional[Iterable[Workspace]] = None, limit: Optional[int] = None):
        if workspaces is None:
            workspaces = []
        else:
            try:
                workspaces = list(workspaces)
            except ValueError as e:
                raise ValueError('Parameter `workspaces` must be iterable: {}'.format(e))

        if limit is None:
            limit = float('+inf')

        self.workspaces = [
            (ws.id, ws)
            for ws in sorted(workspaces, key=attrgetter('id'))
        ]

        self.limit = limit

        if len(self.workspaces) > self.limit:
            warnings.warn('Workspace count is greater than limit')

    @classmethod
    def load_from_directory(self,
                            root,
                            path_structure_cls,
                            data_cls=None,
                            *,
                            workspace_cls=None):
        if workspace_cls is None:
            workspace_cls = Workspace
        workspaces = _load_workspaces(
            root,
            path_structure_cls,
            data_cls,
            workspace_cls,
        )
        return WorkspaceStorage(workspaces)

    def __iter__(self):
        yield from self.keys()

    def keys(self):
        yield from (key for key, _ in self.workspaces)

    def values(self) -> List[Workspace]:
        yield from (value for _, value in self.workspaces)

    def items(self):
        yield from self.workspaces

    def __getitem__(self, item) -> Optional[Workspace]:
        """Gets workspace from storage by id"""
        workspaces = dict(self.workspaces)
        return workspaces[item] if item in workspaces else None

    def at(self, index) -> Optional[Workspace]:
        """Gets workspace from storage by index

        To get the oldest workspace use index 0
        To get the latest workspace use index -1
        """
        if index >= len(self.workspaces):
            return None

        pair = self.workspaces[index]
        return pair[1]

    def add(self, workspace: Workspace):
        """Add a workspace

        Parameters
        ----------
        workspace: subclass of BaseWorkspace
            Workspace to be added
        """
        if not isinstance(workspace, Workspace):
            raise ValueError('Value must be of type {}'.format(Workspace.__name__))

        pair = (workspace.id, workspace)

        workspaces = list(self.workspaces)
        workspaces.append(pair)
        workspaces = sorted(workspaces, key=itemgetter(0))
        self.workspaces = workspaces

    def __repr__(self):
        return '<WorkspaceStorage {}>'.format(list(self.keys()))

    def __len__(self):
        return len(self.workspaces)





class Path(str):
    """String wrapper class to hold a loader attribute"""

    def __new__(cls, path, *args, **kwargs):
        return super().__new__(cls, path)

    def __init__(self, path: str, loader):
        """Creates a path with a loader attribute

        Attributes
        ----------
        path
            Target path
        loader
            Function to be bound to the path access action
        """
        self.loader = loader




from .data_meta import DataMeta


class BaseData(metaclass=DataMeta):
    """Base class for data

    Examples
    --------

    :ref:`basic-data-loader`

    :ref:`pass-data-loader-args`

    :ref:`chained-data-loaders`
    """

    def __init__(self, *args, **kwargs):
        """Creates a Data instance

        Attributes
        ----------
        args
            Will be passed to each loader
        kwargs
            Will be passed to each loader
        """
        self._args = args
        self._kwargs = kwargs



import tempfile
from itertools import starmap
from pathlib import Path

import os
from random import randint
import random
from typing import Iterable

import pytest


def test_manager_latest_workspace():
    from woger import WorkspaceManager, Workspace, BasePathStructure

    class OrderedWorkspace(Workspace):
        def __lt__(self, other):
            return self.state_path < other._state_path

    subdirs = ['134', '235', '9', '17']
    latest_id = 235

    with tempfile.TemporaryDirectory() as root:
        for path in subdirs:
            os.makedirs(str(Path(root) / path), exist_ok=True)

        wm = WorkspaceManager(root, BasePathStructure, workspace_cls=OrderedWorkspace)

        assert wm.latest().id == latest_id


def test_manager_current_and_target():
    from woger import WorkspaceManager, BasePathStructure
    from woger.manager import _TARGET, _CURRENT, _STATE

    names = list(map(str, range(randint(50, 100), randint(950, 1000), 49)))
    random.shuffle(names)
    current, target = names[0], names[-1]
    contents = (
        '{{'
        '"{current_tag}": {current},'
        '"{target_tag}": {target}'
        '}}'
    ).format(
        target=target,
        current=current,
        target_tag=_TARGET,
        current_tag=_CURRENT,
    )

    with tempfile.TemporaryDirectory() as root:
        root = Path(root)
        with open(str(root / _STATE), 'w') as f:
            f.write(contents)

        for name in names:
            os.makedirs(str(root / name), exist_ok=True)

        manager = WorkspaceManager(
            root,
            BasePathStructure,
        )

        assert manager.current().root == root / current
        assert manager.target().root == root / target


_paths = [
    '346',
    '134',
    '576',
]

_contents = [(
    '{"load_raw_xml":"finished","parse_xml": "finished"}',
    '{"load_raw_xml":"finished","parse_xml": "finished"}',
    '{"load_raw_xml": "finished","parse_xml": "started"}',
), (
    '{"load_raw_xml":"finished","parse_xml": "failed"}',
    '{"load_raw_xml": "started"}',
    '{"load_raw_xml": "finished","parse_xml": "started"}',
)]

_checks = [
    lambda ws: ws.id == 346,
    lambda ws: ws is None,
]


class _Meta:
    def __init__(self, path, contents):
        self.path = path
        self.contents = contents


@pytest.mark.parametrize('meta_list,check', [
    (list(starmap(_Meta, zip(_paths, _contents[i]))), _checks[i])
    for i
    in [0, 1]
])
def test_manager_search_latest_finished(meta_list: Iterable[_Meta], check):
    from woger import WorkspaceManager, BasePathStructure, BaseAction
    from woger.manager import _STATE

    with tempfile.TemporaryDirectory() as root:

        root = Path(root)
        for meta in meta_list:
            os.makedirs(str(root / meta.path), exist_ok=True)
            path = root / meta.path / _STATE
            with open(str(path), 'w') as f:
                f.write(meta.contents)

        class Action(BaseAction):
            load_raw_xml = 'load_raw_xml'
            parse_xml = 'parse_xml'

        wm = WorkspaceManager(root, BasePathStructure)
        ws = wm.find_latest_finished(Action.parse_xml)

        print(wm.storage)
        assert check(ws)



import os
from setuptools import find_packages, setup


NAME = 'woger'
DESCRIPTION = 'Workspace manager library'
URL = 'https://gitlab.com/grihabor/woger'
EMAIL = 'grihabor@mail.ru'
AUTHOR = 'Borodin Gregory'

REQUIRED = [
    'pathlib2',
]


def _get_project_path():
    return os.path.abspath(os.path.join(__file__, os.pardir))


def get_version():
    project_path = _get_project_path()
    init_path = os.path.join(project_path, NAME, '__init__.py')
    with open(init_path, 'r') as f:
        for line in f:
            if line.startswith('__version__'):
                version_string = line.split('=')[-1].strip()
                return version_string.strip("'")


def get_readme():
    project_path = _get_project_path()
    readme_path = os.path.join(project_path, 'README.rst')
    with open(readme_path, 'r') as f:
        return f.read()


def main():
    setup(
        name=NAME,
        version=get_version(),
        description=DESCRIPTION,
        author=AUTHOR,
        author_email=EMAIL,
        url=URL,
        packages=find_packages(exclude=('tests',)),
        install_requires=REQUIRED,
        setup_requires=[
            'pytest-runner',
        ],
        tests_require=[
            'pytest',
            'pytest-cov',
        ],
        include_package_data=True,
        license='MIT',
        long_description=get_readme(),
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
        ],
        keywords='woger workspace manager directory structure version control'
    )


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import Union

import os

from .workspace_meta import PathStructureMeta


class BasePathStructure(metaclass=PathStructureMeta):
    def __init__(self, root: Union[bytes, str]):
        """Base class for path structure

        Parameters
        ----------
        root: path-like
            Path structure root


        Examples
        --------

        :ref:`basic-path-structure`

        """
        self._root = Path(root)
        os.makedirs(str(self._root), exist_ok=True)
